from operator import mul
from functools import reduce
from torch.nn import functional as F


def KL_divergence(predict, target):
    return (target * (target / predict).log()).sum() - target.sum() + predict.sum()


def Euclidean(predict, target):
    return F.mse_loss(predict, target, reduction="sum") / 2


def IS_divergence(predict, target):
    div = target / predict
    return div.sum() - div.log().sum() - reduce(mul, target.shape)


def Beta_divergence(predict, target, beta=2):
    if beta == 2:
        return Euclidean(predict, target)
    elif beta == 1:
        return KL_divergence(predict, target)
    elif beta == 0:
        return IS_divergence(predict, target)
    else:
        bminus = beta - 1
        return (
            target.pow(beta).sum()
            + bminus * predict.pow(beta).sum()
            - beta * (target * predict.pow(bminus)).sum()
        ) / (beta * bminus)



import torch
import numpy as np
import scipy.stats as st


def normalize(x: torch.Tensor, axis=0) -> torch.Tensor:
    return x / x.sum(axis, keepdim=True)


def scalar_to_vec(x, k, dist="unif", nsig=3):
    """
    Helper function to generate 1d kernel distributions that integrate
      to a specified scalar. Used in generalizing NMF initialization
      weights to NMFD.
    ------------
    Parameters
    ------------
    x: float to which the 1d kernel should integrate
    k: integer width of the 1d kernel
    dist: probability distribution ("unif" or "gauss") to use for kernel
    nsig: for Gaussian kernels, number of SDs to span in k steps
    ------------
    Returns
    ------------
    numpy array of shape [1,k]
    """
    if dist == "unif":
        return np.ones(k) * (x / k)

    elif dist == "gauss":
        t = np.linspace(-nsig, nsig, k + 1)
        return np.diff(st.norm.cdf(t)) * x

    else:
        raise ValueError('Currently supports only "unif" and "gauss"')


def sweep_components(X, n_max=None, n_min=2):
    """
    Sweeps over all values of n_components and returns a plot of losses vs n_components
    Parameters
    ----------
    X : Tensor
    n_max : int
        Max n in search, default X.shape[0]
    n_min : int
        Min n in search, default 2

    Returns
    -------
    fig, axes
    """
    import matplotlib.pyplot as plt
    from constrainedmf.nmf.models import NMF

    if n_max is None:
        n_max = X.shape[0]

    losses = list()
    kl_losses = list()
    for n_components in range(n_min, n_max + 1):
        nmf = NMF(X.shape, n_components)
        nmf.fit(X, beta=2, tol=1e-8, max_iter=500)
        losses.append(nmf.loss(X, beta=2))
        kl_losses.append(nmf.loss(X, beta=1))

    fig, axes = plt.subplots(1, 2)
    x = list(range(n_min, n_max + 1))
    axes[0].plot(x, losses, label="MSE Loss")
    axes[0].set_title("MSE Loss")
    axes[1].plot(x, kl_losses, label="KL Loss")
    axes[1].set_title("KL Loss")
    fig.suptitle("Loss vs # of Components")
    return fig, axes


def iterative_nmf(
    NMFClass, X, n_components, *, beta=2, alpha=0.0, tol=1e-8, max_iter=1000, **kwargs
):
    """
    Utility for performing NMF on a stream of data along a common state variable
    (Temperature or composition), that coincides with the data ordering.

    Parameters
    ----------
    NMFClass : class
        Child class of NMFBase
    X : Tensor
        Data to perform NMF on
    n_components : int
        Number of components for NMF
    beta : int
        Beta for determining loss function
    alpha : float
        Alpha for determining regularization. Default 0.0 is no regularization.
    tol : float
        Optimization tolerance
    max_iter : int
        Maximum optimization iterations
    kwargs : dict
        Passed to intialization of NMF

    Returns
    -------
    nmfs : list of NMF instances

    """
    nmfs = list()
    initial_components = [torch.rand(1, X.shape[-1]) for _ in range(n_components)]
    fix_components = [False for _ in range(n_components)]

    # Start on bounding the outer components
    initial_components[0] = X[0, :].reshape(1, -1)
    fix_components[0] = True
    initial_components[-1] = X[-1, :].reshape(1, -1)
    fix_components[-1] = True

    nmf = NMFClass(
        X.shape,
        n_components,
        initial_components=initial_components,
        fix_components=fix_components,
        **kwargs
    )
    nmf.fit(X, beta=beta, tol=tol, max_iter=max_iter, alpha=alpha)
    nmfs.append(nmf)

    if len(nmf.W.shape) == 3:
        convolutional = True
    else:
        convolutional = False
    visited = {0, n_components - 1}
    for _ in range(n_components - 2):

        # Find next most prominent weight
        if convolutional:
            indices = (
                nmf.W.sum(axis=-1).max(axis=0).values.argsort(descending=True).numpy()
            )
        else:
            indices = nmf.W.max(axis=0).values.argsort(descending=True).numpy()
        for i in indices:
            if i in visited:
                continue
            else:
                visited.add(i)
                weight_idx = i
                break

        # Find most important component to that weight
        if convolutional:
            pattern_idx = int(nmf.W.sum(axis=-1).argmax(axis=0)[weight_idx])
        else:
            pattern_idx = int(nmf.W.argmax(axis=0)[weight_idx])

        # Lock and run
        initial_components[weight_idx] = X[pattern_idx, :].reshape(1, -1)
        fix_components[weight_idx] = True
        nmf = NMFClass(
            X.shape,
            n_components,
            initial_components=initial_components,
            fix_components=fix_components,
            **kwargs
        )
        nmf.fit(X, beta=beta, tol=tol, max_iter=max_iter, alpha=alpha)
        nmfs.append(nmf)

    return nmfs



from constrainedmf.nmf.models import NMF
import numpy as np
import torch
from pathlib import Path
from constrainedmf.companion.plotting import decomp_plot
from constrainedmf.nmf.utils import sweep_components

torch.manual_seed(1234)
np.random.seed(1234)


def get_data(data_dir=None):
    if data_dir is None:
        data_dir = Path("~/Datasets/HTPXRD/BaTiO/exp_patterns").expanduser()
    if data_dir is None:
        data_dir = Path(__file__).parents[1] / "example_data/NaCl_CrCl3_pdf_ramp"
    paths = sorted(list(data_dir.glob("*.npy")))
    profiles = list()
    T = list()
    for path in paths:
        x = np.load(path)
        T.append(
            float(str(path).split("_")[2][:3])
        )  # hardcoded nonsense for T from label
        profiles.append(x / x.max())
    X = torch.tensor(np.concatenate(profiles, axis=1).T, dtype=torch.float)
    T = np.array(T)
    return T, X


def standard_nmf(X):
    nmf = NMF(X.shape, 4)
    nmf.fit(X, beta=2, tol=1e-8, max_iter=1000)
    return nmf


def plot_adjustments(axes):
    for ax in axes[1:]:
        ax.axvline(185, linestyle="--", color="k")
        ax.axvline(280, linestyle="--", color="k")
        ax.axvline(400, linestyle="--", color="k")
        ax.set_ylim(0, 1)
        ax.set_yticks([0, 0.25, 0.5, 0.75, 1.0])
        ax.set_yticklabels([0, 0.25, 0.5, 0.75, 1])
        ax.set_xlabel("Temperature [K]")
        ax.set_ylabel("$x_\Phi$")  # noqa: W605
        ax.set_xlim(150, 445)


def make_plots():
    from constrainedmf.nmf.utils import iterative_nmf

    T, X = get_data()
    figs = list()
    axes = list()
    fig, ax = sweep_components(X, n_max=8)
    figs.append(fig),
    axes.append(ax)
    fig, ax = decomp_plot(standard_nmf(X), T)
    plot_adjustments(ax)
    figs.append(fig)
    axes.append(ax)
    for nmf in iterative_nmf(NMF, X, n_components=4, beta=2, tol=1e-8, max_iter=1000):
        fig, ax = decomp_plot(nmf, T)
        plot_adjustments(ax)
        figs.append(fig)
        axes.append(ax)

    return figs


if __name__ == "__main__":
    path = Path(__file__).parent / "example_output"
    path.mkdir(exist_ok=True)
    figs = make_plots()
    for i, fig in enumerate(figs):
        fig.tight_layout()
        fig.show()
        fig.savefig(path / f"BaTiO_{i}.png")



import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.gridspec import GridSpec
import numpy as np
import torch


def toy_plot(model, x, Y, weights, components):
    fig = plt.figure(tight_layout=True, figsize=(15, 10))
    gs = GridSpec(2, 6)
    recon_axes = [fig.add_subplot(gs[0, i * 2 : (i + 1) * 2]) for i in range(3)]
    comp_ax = fig.add_subplot(gs[1, :3])
    weight_ax = fig.add_subplot(gs[1, 3:])

    with torch.no_grad():
        recon = model.reconstruct(model.H, model.W)

    ax = recon_axes[0]
    ax.plot(x, Y[0, :], label="Truth")
    ax.plot(x, recon[0, :].data.numpy(), label="Reconstruct")
    ax = recon_axes[1]
    ax.plot(x, Y[Y.shape[0] // 2, :], label="Truth")
    ax.plot(x, recon[recon.shape[0] // 2, :].data.numpy(), label="Reconstruct")
    ax = recon_axes[2]
    ax.plot(x, Y[-1, :], label="Truth")
    ax.plot(x, recon[-1, :].data.numpy(), label="Reconstruct")
    for ax in recon_axes:
        ax.legend()

    H = model.H.data.numpy()
    ax = comp_ax
    for i in range(H.shape[0]):
        ax.plot(x, H[i, :], label=f"Learned Component {i}")
        ax.plot(x, components[i, :], "--k", label=f"True Component {i}")
    ax.set_title("Learned Components")
    ax.legend()

    W = model.W.data.numpy()
    ax = weight_ax
    for i in range(W.shape[1]):
        ax.plot(W[:, i], label=f"Learned Weights {i}")
        ax.plot(weights.T[:, i], "--k", label=f"True Weights {i}")
    ax.set_title("Learned Weights")
    ax.legend()
    return fig


def decomp_plot(nmf, T, axes=None):
    H = nmf.H.data.numpy()
    W = nmf.W.data.numpy()
    fig, axes = plt.subplots(1, 3, figsize=(15, 4))
    ax = axes[0]
    for i in range(H.shape[0]):
        ax.plot(H[i, :] / H[i, :].max() + i)
    ax.set_title("Stacked Normalized Components")

    ax = axes[1]
    for i in range(W.shape[1]):
        ax.plot(T, W[:, i])
    ax.set_title("Weights")

    ax = axes[2]
    for i in range(W.shape[1]):
        ax.plot(T, W[:, i] / W[:, i].max())
    ax.set_title("Normalized Weights")
    return fig, axes


def waterfall(ax, xs, ys, alphas, color="k", sampling=1, offset=0.2, **kwargs):
    indicies = range(0, xs.shape[0])[::sampling]
    for plt_i, idx in enumerate(indicies):
        y = ys[idx, :] + plt_i * offset
        x = xs[idx, :]
        ax.plot(x, y, color=color, alpha=alphas[idx], **kwargs)
    return ax


def summary_plot(
    sub_Q,
    sub_I,
    alphas,
    axes=None,
    sax=None,
    components=None,
    comax=None,
    cmap="tab10",
    alt_ordinate=None,
    offset=1.0,
    summary_fig=False,
):
    """
    Example plotting of NMF results. Not necessarily for Bluesky deployment

    Parameters
    ----------
    sub_Q: array
        Q to plot in I(Q)
    sub_I: array
        I to plot in I(Q)
    alphas: array
        transparencies of multiple repeated plots of I(Q)
    axes: optional existing axes for waterfalls
    sax: optional axes for summary figure
    cmap: mpl colormap
    alt_ordinate: array
        Array len sub_I.shape[0], corresponding to an alternative labeled dimension for which to order the stacked plots
    summary_fig: bool
        Whether to include separate figure of alphas over the ordinate

    Returns
    -------
    fig, axes

    """

    n_components = alphas.shape[1]
    cmap = mpl.cm.get_cmap(cmap)
    norm = mpl.colors.Normalize(vmin=0, vmax=n_components)

    # Create alternative ordinate for the waterfall/stacking
    if alt_ordinate is not None:
        idxs, labels = list(
            zip(*sorted(zip(range(sub_I.shape[0]), alt_ordinate), key=lambda x: x[1]))
        )
    else:
        idxs = list(range(sub_I.shape[0]))

    xs = sub_Q[idxs, :]
    ys = sub_I[idxs, :]
    alphas = alphas[idxs, :]

    # Order by proxy center of mass of class in plot regime. Makes the plots feel like a progression not random.
    alpha_ord = np.argsort(np.matmul(np.arange(alphas.shape[0]), alphas))

    if axes is None:
        fig, axes = plt.subplots(
            int(np.ceil(np.sqrt(n_components))), int(np.ceil(np.sqrt(n_components)))
        )
        axes = axes.reshape(-1)
    else:
        axes = np.ravel(axes)
    for i, ax in enumerate(axes):
        if i < n_components:
            i_a = alpha_ord[i]
            color = cmap(norm(i))
            alpha = (alphas[:, i_a] - np.min(alphas[:, i_a])) / (
                np.max(alphas[:, i_a]) - np.min(alphas[:, i_a])
            )
            waterfall(ax, xs, ys, alpha, color=color, offset=offset)
        else:
            ax.set_visible = False

    if summary_fig:
        if sax is None:
            sfig, sax = plt.subplots(figsize=(6, 6))

        sx = np.arange(0, alphas.shape[0])
        for i in range(alphas.shape[1]):
            sax.plot(
                sx,
                alphas[:, alpha_ord[i]],
                color=cmap(norm(i)),
                label=f"Component {i + 1}",
            )

    if components is not None:
        if comax is None:
            comfig, comax = plt.subplots(figsize=(6, 6))
        for i in range(components.shape[0]):
            kernel_width = xs.shape[1] - components.shape[1] + 1
            if kernel_width == 1:
                comax.plot(
                    xs[0][:],
                    components[alpha_ord[i], :] + i,
                    color=cmap(norm(i)),
                )
            else:
                comax.plot(
                    xs[0][kernel_width // 2 : -kernel_width // 2 + 1],
                    components[alpha_ord[i], :] + i,
                    color=cmap(norm(i)),
                )

    return



"Non-negative matrix factorization of full datasets"
from constrainedmf.nmf.models import NMF, NMFD
from constrainedmf.nmf.utils import iterative_nmf
import torch
import numpy as np


def decomposition(
    Q,
    I,
    n_components=3,
    q_range=None,
    initial_components=None,
    fix_components=(),
    mode="Linear",
    kernel_width=1,
    max_iter=1000,
    bkg_removal=None,
    normalize=False,
    **kwargs,
):
    """
    Decompose and label a set of I(Q) data with optional focus bounds. Can be used for other
    1-dimensional response functions, written with total scattering in mind.

    Two operating modes are available: Linear (conventional) and Deconvolutional. The former will proceed as conventional
    NMF as implemented in sklearn, with the added flexibility of the torch implementation. The latter will include a
    convolutional kernel in the reconstruction between the component and weight matricies.

    Initial components can be set as starting conditions of the component matrix for the optimization. These components
    can be fixed or allowed to vary using the `fix_components` argument as a tuple of booleans.

    Keyword arguments are passed to the fit method

    Parameters
    ----------
    Q : array
        Ordinate Q for I(Q). Assumed to be rank 2, shape (m_patterns, n_data)
    I : array
        The intensity values for each Q, assumed to be the same shape as Q. (m_patterns, n_data)
    n_components: int
        Number of components for NMF
    q_range : tuple, list
        (Min, Max) Q values for consideration in NMF. This enables a focused region for decomposition.
    initial_components: array
        Initial starting conditions of intensity components. Assumed to be shape (n_components, n_data).
        If q_range is given, these will be trimmed in accordance with I.
    fix_components: tuple(bool)
        Flags for fixing a subset of initial components
    mode: {"Linear", "Deconvolutional"}
        Operating mode
    kernel_width: int
        Width of 1-dimensional convolutional kernel
    max_iter: int
        Maximum number of iterations for NMF
    bkg_removal: int, optional
        Integer degree for peakutils background removal
    normalize: bool, optional
        Flag for min-max normalization
    **kwargs: dict
        Arguments passed to the fit method. See nmf.models.NMFBase.

    Returns
    -------
    sub_Q : array
        Subsampled ordinate used for NMF
    sub_I : array
        Subsampled I used for NMF
    alphas : array
        Resultant weights from NMF
    components:  array
        Resultant components from NMF

    """

    # Data subselection
    if q_range is None:
        idx_min = 0
        idx_max = I.shape[1]
    else:
        idx_min = (
            np.where(Q[0, :] < q_range[0])[0][-1]
            if len(np.where(Q[0, :] < q_range[0])[0])
            else 0
        )
        idx_max = (
            np.where(Q[0, :] > q_range[1])[0][0]
            if len(np.where(Q[0, :] > q_range[1])[0])
            else I.shape[1]
        )

    sub_I = I[:, idx_min:idx_max]
    sub_Q = Q[:, idx_min:idx_max]

    # Data manipulation
    if bkg_removal:
        import peakutils

        bases = []
        for i in range(sub_I.shape[0]):
            bases.append(peakutils.baseline(sub_I[i, :], deg=bkg_removal))
        bases = np.stack(bases)
        sub_I = sub_I - bases
    if normalize:
        sub_I = (sub_I - np.min(sub_I, axis=1, keepdims=True)) / (
            np.max(sub_I, axis=1, keepdims=True) - np.min(sub_I, axis=1, keepdims=True)
        )

    # Numerical stability of non-negativity
    if np.min(sub_I) < 0:
        sub_I = sub_I - np.min(sub_I, axis=1, keepdims=True)

    # Initial components
    if mode != "Deconvolutional":
        kernel_width = 1
    n_features = sub_I.shape[1]
    if initial_components is None:
        input_H = None
    else:
        input_H = []
        for i in range(n_components):
            try:
                sub_H = initial_components[i][idx_min:idx_max]
                sub_H = sub_H[kernel_width // 2 : len(sub_H) - kernel_width // 2 + 1]
                if normalize:
                    sub_H = (sub_H - np.min(sub_H)) / (np.max(sub_H) - np.min(sub_H))
                input_H.append(
                    torch.tensor(sub_H, dtype=torch.float).reshape(
                        1, n_features - kernel_width + 1
                    )
                )
            except IndexError:
                input_H.append(torch.rand(1, n_features - kernel_width + 1))

    # Model construction
    if mode == "Linear":
        model = NMF(
            sub_I.shape,
            n_components,
            initial_components=input_H,
            fix_components=fix_components,
        )
    elif mode == "Deconvolutional":
        model = NMFD(
            sub_I.shape,
            n_components,
            T=kernel_width,
            initial_components=input_H,
            fix_components=fix_components,
        )
    else:
        raise NotImplementedError

    _, W = model.fit_transform(torch.tensor(sub_I), max_iter=max_iter, **kwargs)

    if len(W.shape) > 2:
        alphas = torch.mean(W, 2).data.numpy()
    else:
        alphas = W.data.numpy()

    components = torch.cat([x for x in model.H_list]).data.numpy()
    return sub_Q, sub_I, alphas, components


def iterative_decomposition(
    Q,
    I,
    n_components=3,
    q_range=None,
    mode="Linear",
    kernel_width=1,
    max_iter=1000,
    bkg_removal=None,
    normalize=False,
    **kwargs,
):
    # Data subselection
    if q_range is None:
        idx_min = 0
        idx_max = I.shape[1]
    else:
        idx_min = (
            np.where(Q[0, :] < q_range[0])[0][-1]
            if len(np.where(Q[0, :] < q_range[0])[0])
            else 0
        )
        idx_max = (
            np.where(Q[0, :] > q_range[1])[0][0]
            if len(np.where(Q[0, :] > q_range[1])[0])
            else I.shape[1]
        )

    sub_I = I[:, idx_min:idx_max]
    sub_Q = Q[:, idx_min:idx_max]

    # Data manipulation
    if bkg_removal:
        import peakutils

        bases = []
        for i in range(sub_I.shape[0]):
            bases.append(peakutils.baseline(sub_I[i, :], deg=bkg_removal))
        bases = np.stack(bases)
        sub_I = sub_I - bases
    if normalize:
        sub_I = (sub_I - np.min(sub_I, axis=1, keepdims=True)) / (
            np.max(sub_I, axis=1, keepdims=True) - np.min(sub_I, axis=1, keepdims=True)
        )

    # Numerical stability of non-negativity
    if np.min(sub_I) < 0:
        sub_I = sub_I - np.min(sub_I, axis=1, keepdims=True)

    if mode == "Deconvolutional":
        nmf_class = NMFD
    elif mode == "Linear":
        nmf_class = NMF
    else:
        raise NotImplementedError(f"Mode {mode} unavailable.")

    # Safety check
    if "initial_components" in kwargs:
        del kwargs["initial_components"]
    if "fix_components" in kwargs:
        del kwargs["fix_components"]

    nmfs = iterative_nmf(
        nmf_class,
        torch.tensor(sub_I, dtype=torch.float),
        n_components=n_components,
        max_iter=max_iter,
        kernel_width=kernel_width,
        **kwargs,
    )

    nmf = nmfs[-1]
    W = nmf.W
    H = nmf.H
    if len(W.shape) > 2:
        alphas = torch.mean(W, 2).data.numpy()
    else:
        alphas = W.data.numpy()
    components = H.data.numpy()
    return sub_Q, sub_I, alphas, components



import torch
import torch.nn as nn
import torch.nn.functional as F
from constrainedmf.nmf.metrics import Beta_divergence


def _mu_update(param, pos, gamma, l1_reg, l2_reg):
    if isinstance(param, nn.ParameterList):
        # Handle no gradients in fixed components
        grad = torch.cat(
            [x.grad if x.requires_grad else torch.zeros_like(x) for x in param]
        )
    elif param.grad is None:
        return
    else:
        grad = param.grad
    # prevent negative term, very likely to happen with kl divergence
    multiplier = F.relu(pos - grad, inplace=True)

    if l1_reg > 0:
        pos.add_(l1_reg)
    if l2_reg > 0:
        if isinstance(param, nn.ParameterList):
            reg_param = torch.cat([x for x in param])
        else:
            reg_param = param
        if pos.shape != reg_param.shape:
            pos = pos + l2_reg * reg_param
        else:
            pos.add_(l2_reg * reg_param)

    multiplier.div_(pos)
    if gamma != 1:
        multiplier.pow_(gamma)
    if isinstance(param, nn.ParameterList):
        for i, sub_param in enumerate(param):
            sub_param.mul_(multiplier[i, :])
    else:
        param.mul_(multiplier)


class NMFBase(nn.Module):
    def __init__(
        self,
        W_shape,
        H_shape,
        n_components,
        initial_components=None,
        fix_components=(),
        initial_weights=None,
        fix_weights=(),
        **kwargs
    ):
        """
        Base class for setting up NMF
        W is (m_examples, n_components) or (m_examples, n_components, kernel_width)
        H is (n_components, n_example_features)
        W @ H give reconstruction of X.
        Parameters
        ----------
        W_shape
        H_shape
        n_components
        initial_components
        fix_components
        """
        super().__init__()
        self.fix_neg = nn.Threshold(0.0, 1e-8)
        self.rank = n_components

        W = torch.rand(*W_shape)
        self.W = nn.Parameter(W)
        if initial_weights is not None:
            w_list = [nn.Parameter(weight) for weight in initial_weights]
        else:
            w_list = [
                nn.Parameter(torch.rand(1, *W_shape[1:])) for _ in range(W_shape[0])
            ]
        if fix_weights:
            for i in range(len(fix_weights)):
                w_list[i].requires_grad = not fix_weights[i]
        self.W_list = nn.ParameterList(w_list)

        if initial_components is not None:
            h_list = [nn.Parameter(component) for component in initial_components]
        else:
            h_list = [
                nn.Parameter(torch.rand(1, *H_shape[1:])) for _ in range(H_shape[0])
            ]
        if fix_components:
            for i in range(len(fix_components)):
                h_list[i].requires_grad = not fix_components[i]
        self.H_list = nn.ParameterList(h_list)

    @property
    def H(self):
        return torch.cat([x for x in self.H_list])

    @property
    def W(self):
        return torch.cat([x for x in self.W_list])

    def loss(self, X, beta=2):
        with torch.no_grad():
            WH = self.reconstruct(self.H, self.W)
            return Beta_divergence(self.fix_neg(WH), X, beta)

    def forward(self, H=None, W=None):
        if H is None:
            H = self.H
        if W is None:
            W = self.W
        return self.reconstruct(H, W)

    def reconstruct(self, H, W):
        raise NotImplementedError

    def get_W_positive(self, WH, beta, H_sum) -> (torch.Tensor, None or torch.Tensor):
        raise NotImplementedError

    def get_H_positive(self, WH, beta, W_sum) -> (torch.Tensor, None or torch.Tensor):
        raise NotImplementedError

    def fit(
        self,
        X,
        update_W=True,
        update_H=True,
        beta=1,
        tol=1e-5,
        max_iter=200,
        alpha=0,
        l1_ratio=0,
    ):

        X = X.type(torch.float)
        X = self.fix_neg(X)
        # self.W.requires_grad = update_W
        # self.H.requires_grad = update_H

        if beta < 1:
            gamma = 1 / (2 - beta)
        elif beta > 2:
            gamma = 1 / (beta - 1)
        else:
            gamma = 1

        l1_reg = alpha * l1_ratio
        l2_reg = alpha * (1 - l1_ratio)

        loss_scale = torch.prod(torch.tensor(X.shape)).float()

        H_sum, W_sum = None, None

        for n_iter in range(max_iter):
            # W update
            if update_W and any([x.requires_grad for x in self.W_list]):
                self.zero_grad()
                WH = self.reconstruct(self.H.detach(), self.W)
                loss = Beta_divergence(self.fix_neg(WH), X, beta)
                loss.backward()

                with torch.no_grad():
                    positive_comps, H_sum = self.get_W_positive(WH, beta, H_sum)
                    _mu_update(self.W_list, positive_comps, gamma, l1_reg, l2_reg)
                W_sum = None

            # H update
            if update_H and any([x.requires_grad for x in self.H_list]):
                self.zero_grad()
                WH = self.reconstruct(self.H, self.W.detach())
                loss = Beta_divergence(self.fix_neg(WH), X, beta)
                loss.backward()

                with torch.no_grad():
                    positive_comps, W_sum = self.get_H_positive(WH, beta, W_sum)
                    _mu_update(self.H_list, positive_comps, gamma, l1_reg, l2_reg)
                H_sum = None

            loss = loss.div_(loss_scale).item()

            if not n_iter:
                loss_init = loss
            elif (previous_loss - loss) / loss_init < tol:  # noqa: F821
                break
            previous_loss = loss  # noqa:F841

        return n_iter

    def fit_transform(self, *args, **kwargs):
        n_iter = self.fit(*args, **kwargs)
        return n_iter, torch.cat([x for x in self.W_list])


class NMF(NMFBase):
    def __init__(self, X_shape, n_components, **kwargs):
        self.m_examples, self.n_features = X_shape
        super().__init__(
            (self.m_examples, n_components),
            (n_components, self.n_features),
            n_components,
            **kwargs
        )

    def reconstruct(self, H, W):
        return W @ H

    def get_W_positive(self, WH, beta, H_sum):
        H = self.H
        if beta == 1:
            if H_sum is None:
                H_sum = H.sum(1)
            denominator = H_sum[None, :]
        else:
            if beta != 2:
                WH = WH.pow(beta - 1)
            WHHt = WH @ H.t()
            denominator = WHHt

        return denominator, H_sum

    def get_H_positive(self, WH, beta, W_sum):
        W = self.W
        if beta == 1:
            if W_sum is None:
                W_sum = W.sum(0)  # shape(n_components, )
            denominator = W_sum[:, None]
        else:
            if beta != 2:
                WH = WH.pow(beta - 1)
            WtWH = W.t() @ WH
            denominator = WtWH
        return denominator, W_sum

    def sort(self):
        raise NotImplementedError


class NMFD(NMFBase):
    def __init__(self, X_shape, n_components, T=1, **kwargs):
        self.m_examples, self.n_features = X_shape
        self.pad_size = T - 1
        super().__init__(
            (self.m_examples, n_components, T),
            (n_components, self.n_features - T + 1),
            n_components,
            **kwargs
        )

    def reconstruct(self, H, W):
        return F.conv1d(H[None, :], W.flip(2), padding=self.pad_size)[0]

    def get_W_positive(self, WH, beta, H_sum):
        H = self.H
        if beta == 1:
            if H_sum is None:
                H_sum = H.sum(1)
            denominator = H_sum[None, :, None]
        else:
            if beta != 2:
                WH = WH.pow(beta - 1)
            WHHt = F.conv1d(WH[:, None], H[:, None])
            denominator = WHHt

        return denominator, H_sum

    def get_H_positive(self, WH, beta, W_sum):
        W = self.W
        if beta == 1:
            if W_sum is None:
                W_sum = W.sum((0, 2))
            denominator = W_sum[:, None]
        else:
            if beta != 2:
                WH = WH.pow(beta - 1)
            WtWH = F.conv1d(WH[None, :], W.transpose(0, 1))[0]
            denominator = WtWH
        return denominator, W_sum

    def sort(self):
        raise NotImplementedError



from constrainedmf.nmf.models import NMF
import numpy as np
import torch
from constrainedmf.companion.plotting import toy_plot

torch.manual_seed(1234)
np.random.seed(1234)


def gaussian(x, mu, sig, a):
    return a * np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))


def tophat(x, s1, s2, a):
    return a * (np.heaviside(x - s1, 1) - np.heaviside(x - s2, 1))


def lorentzian(x, x0, gam, a):
    return a * gam ** 2 / (gam ** 2 + (x - x0) ** 2)


def norm_poly_weights(x, a0, a1, a2):
    this_max = np.max(a0 + x * a1 + a2 * x ** 2)
    this_min = np.min(a0 + x * a1 + a2 * x ** 2)
    return np.nan_to_num(
        ((a0 + x * a1 + a2 * x ** 2) - this_min) / (this_max - this_min)
    )


def poly_weights(x, a0, a1, a2):
    return a0 + x * a1 + a2 * x ** 2


def sigmoid(x, c, p=1):
    return 1.0 / (1.0 + np.exp(-(x - c)) ** p)


def weight_func(t):
    f1 = norm_poly_weights(t, 4, 1, -2)  # * gauss(x,5, .5, 2)
    f2 = poly_weights(t, 0.1, 0.02, -0.0002)  # * tophat(x, 3, 7, 2)
    f3 = sigmoid(t, 70, p=0.5)  # * lorentzian(x, 5, .1, 2)
    return f1, f2, f3


def merge_func(x, weights, t):
    f = weights[0][t] * gaussian(x, 6, 0.5, 2)
    f += weights[1][t] * tophat(x, 4, 6, 2)
    f += weights[2][t] * lorentzian(x, 3, 0.1, 1)
    f += np.random.normal(0, 0.01, f.shape)
    return f


def construct_overlap(n_features=1000, m_patterns=100):
    x = np.linspace(0, 10, n_features)
    y1 = gaussian(x, 6, 0.5, 2)
    y2 = tophat(x, 4, 6, 2)
    y3 = lorentzian(x, 3, 0.1, 1)
    components = np.stack([y1, y2, y3])
    t = np.arange(0, m_patterns, 1)
    weights = np.array(list(weight_func(t)))
    Y = np.matmul(weights.T, components) + np.random.normal(
        0, 0.005, (m_patterns, n_features)
    )
    return x, Y, weights, components


def standard_nmf(X):
    nmf = NMF(X.shape, n_components=3)
    nmf.fit(torch.tensor(X), beta=2)
    return nmf


def constrained_nmf(X, components):
    input_H = [
        torch.tensor(component[None, :], dtype=torch.float) for component in components
    ]
    nmf = NMF(
        X.shape,
        n_components=3,
        initial_components=input_H,
        fix_components=[True for _ in range(len(input_H))],
    )
    nmf.fit(torch.tensor(X), beta=2)
    return nmf


def model_plot(x_fun, model_fun):
    """Takes two callables, one to gen data, and one to train model on data"""

    x, Y, weights, components = x_fun()
    model = model_fun(Y, components)
    fig = toy_plot(model, x, Y, weights, components)
    return fig


def compare_figs():
    fig1 = model_plot(lambda: construct_overlap(), lambda x, comp: standard_nmf(x))
    fig2 = model_plot(
        lambda: construct_overlap(), lambda x, comp: constrained_nmf(x, comp)
    )
    return fig1, fig2


if __name__ == "__main__":
    from pathlib import Path

    path = Path(__file__).parent / "example_output"
    path.mkdir(exist_ok=True)
    fig1, fig2 = compare_figs()
    fig1.savefig(path / "standard_components.png")
    fig2.savefig(path / "constrained_components.png")



from constrainedmf.nmf.models import NMF
import numpy as np
import torch
from constrainedmf.companion.plotting import toy_plot

torch.manual_seed(1234)
np.random.seed(1234)


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))


def construct_overlap(n_features=1000, m_patterns=21, mu=1.2, sig=1):
    x = np.linspace(-10, 10, n_features)
    y1 = gaussian(x, mu, sig)
    y2 = gaussian(x, -1 * mu, sig)
    weights = np.array(list(zip(*[(x, 1 - x) for x in np.linspace(0, 1, m_patterns)])))
    weights += np.random.normal(0, 0.01, weights.shape)
    Y = np.matmul(weights.T, np.array([y1, y2])) + np.random.normal(
        0, 0.01, (m_patterns, n_features)
    )
    return x, Y, weights, np.stack([y1, y2], axis=0)


def standard_nmf(X):
    nmf = NMF(X.shape, n_components=2)
    nmf.fit(torch.tensor(X), beta=2)
    return nmf


def constrained_nmf(X):
    ideal_weights = np.array(
        list(zip(*[(x, 1 - x) for x in np.linspace(0, 1, X.shape[0])]))
    )
    input_W = [torch.tensor(w[None, :], dtype=torch.float) for w in ideal_weights.T]
    nmf = NMF(
        X.shape,
        n_components=2,
        initial_weights=input_W,
        fix_weights=[True for _ in range(len(input_W))],
    )
    nmf.fit(torch.tensor(X), beta=2)
    return nmf


def model_plot(x_fun, model_fun):
    """Takes two callables, one to gen data, and one to train model on data"""

    x, Y, weights, components = x_fun()
    model = model_fun(Y)
    fig = toy_plot(model, x, Y, weights, components)
    return fig


def compare_figs():
    fig1 = model_plot(lambda: construct_overlap(), standard_nmf)
    fig2 = model_plot(lambda: construct_overlap(), constrained_nmf)
    return fig1, fig2


if __name__ == "__main__":
    from pathlib import Path

    path = Path(__file__).parent / "example_output"
    path.mkdir(exist_ok=True)
    fig1, fig2 = compare_figs()
    fig1.savefig(path / "standard_weights.png")
    fig2.savefig(path / "constrained_weights.png")



from pathlib import Path
import numpy as np
from time import sleep, time
import matplotlib.pyplot as plt
from constrainedmf.companion.nmf import decomposition, iterative_decomposition
from constrainedmf.companion.plotting import summary_plot
from IPython import display


class DirectoryAgent:
    def __init__(
        self,
        data_dir,
        n_components,
        *,
        data_spec=None,
        x_lim=None,
        component_dir=None,
        output_dir=None,
        header=0,
        file_ordering=None,
        file_limit=None,
        figsize=None,
        **kwargs,
    ):
        """
        Class for building trained model and classifying a directory.
        Classification can be accomplished on the fly or once using the spin() method.

        Parameters
        ----------
        data_dir: pathlike
            Directory containing data to be classified
        n_components:  int
            Number of components for NMF
        data_spec: basestring
            String specification for glob() method in searching data_dir
            Default behavior is to include all files. If writing temporary files, it is important to include
            final file spec such as '*.xy'.
        x_lim:  tuple
            Size two tuple for bounding the Xs to a region of interest
        component_dir: pathlike
            Directory containing initial components.
        header: int
            Number of header lines in file
        training_output_dir: pathlike
            Output directory of training containing checkpoints for loading
        path_to_model: pathlike
            path to model to load in full (not presently implemented)
        file_ordering: function
            Function for sorting file paths as key argument in sorted
        file_limit: int
            Maximum number of files to consider
        figsize: tuple
            Two integer tuple for matplotlib figsize. Keep in mind all plots appear in a row.
        **kwargs:
            Keyword arguments to pass to companion.nmf.decomposition
        """

        self.dir = Path(data_dir).expanduser()
        if component_dir is None:
            self.component_dir = None
        else:
            self.component_dir = Path(component_dir).expanduser()
        self.n_components = n_components
        if data_spec is None:
            self.path_spec = "*"
        else:
            self.path_spec = data_spec
        self.output_dir = output_dir
        self.paths = []
        self.Ys = []
        self.Xs = []
        self.initial_components = []
        self.limit = file_limit
        self.x_lim = x_lim
        self.header = header
        self.fig = plt.figure(figsize=figsize)
        self.decomposition_args = kwargs
        self.components = None
        self.weights = None

        # TODO: Elegantly understand if I'm in a GUI or in ipython inline
        try:
            self.fig.canvas.manager.show()
        except:  # noqa: E722
            display.display(self.fig)

        if file_ordering is None:
            self.file_ordering = lambda x: x
        else:
            self.file_ordering = file_ordering

    def __len__(self):
        return len(self.paths)

    def path_list(self):
        return sorted(list(self.dir.glob(self.path_spec)), key=self.file_ordering)

    def load_files(self, paths):
        xs = []
        ys = []
        paths = sorted(paths, key=self.file_ordering)
        for idx, path in enumerate(paths):
            if not (self.limit is None) and idx >= self.limit:
                break
            _x, _y = np.loadtxt(path, comments="#", skiprows=self.header).T
            xs.append(_x)
            ys.append(_y)
        return xs, ys

    def _decomposition(self, *args, **kwargs):
        return decomposition(*args, **kwargs)

    def update_plot(self):
        if len(self) < 2:
            return
        Xs = np.array(self.Xs)
        Ys = np.array(self.Ys)
        sub_X, sub_Y, alphas, components = self._decomposition(
            Xs,
            Ys,
            q_range=self.x_lim,
            n_components=self.n_components,
            normalize=True,
            initial_components=self.initial_components,
            fix_components=[True for _ in range(len(self.initial_components))],
            **self.decomposition_args,
        )

        self.components = components
        self.weights = alphas

        self.fig.clf()
        axes = self.fig.subplots(1, self.n_components + 2)
        summary_plot(
            sub_X,
            sub_Y,
            alphas,
            axes=axes[:-1],
            sax=axes[-2],
            components=components,
            comax=axes[-1],
            summary_fig=True,
        )

        # TODO: Elegantly understand if I'm in a GUI or in ipython inline
        self.fig.patch.set_facecolor("white")
        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()
        display.clear_output(wait=True)
        display.display(self.fig)

    def spin(self, sleep_delay=60, verbose=False, timeout=0):
        """
        Starts the spin to read new files and append their classifications to the output and internal dictionary
        If a single pass of the data_directory is required, use a short or negative timeout time.
        This can be run as a multiprocessing.Process target with a Manager
        to retain the output list if being run dynamically.

        Parameters
        ----------
        sleep_delay: float
            number of seconds to wait before checking directory for new files
        verbose: bool
            Print classifications to screen
        timeout: float
            Time to wait before stop spinning. Default is infinite spin.
            If a single pass is required, use a negative value.

        Returns
        -------
        self.classifications:
            dictionary of file basenames and classifications

        """
        start_time = time()

        while True:
            if len(self.path_list()) != len(self):
                self.fig.clf()
                for path in self.path_list():
                    if path.name not in self.paths:
                        self.paths.append(path.name)
                        xs, ys = self.load_files([path])
                        self.Xs.extend(xs)
                        self.Ys.extend(ys)
            if self.component_dir:
                _, self.initial_components = self.load_files(
                    list(self.component_dir.glob(self.path_spec))
                )
            if len(self.initial_components) > self.n_components:
                raise RuntimeError(
                    f"Found {len(self.initial_components)} files for initial components, "
                    f"but the NMF is set to use only {self.n_components} components"
                )
            self.update_plot()
            if timeout and time() - start_time > timeout:
                break
            sleep(sleep_delay)
        return np.array(self.Xs), np.array(self.Ys)

    def elbow_plot(self, n_max):
        """
        Sweeps over all values of n_components and returns a plot of losses vs n_components
        Parameters
        ----------
        n_max : int
            Max n in search, default X.shape[0]

        Returns
        -------
        fig, axes
        """
        import torch
        from constrainedmf.nmf.utils import sweep_components

        xs, ys = self.load_files(self.path_list())
        Y = torch.tensor(ys)
        return sweep_components(Y, n_max=n_max)


class AutoDirectoryAgent(DirectoryAgent):
    def __init__(
        self,
        data_dir,
        n_components,
        *,
        data_spec=None,
        x_lim=None,
        output_dir=None,
        header=0,
        file_ordering=None,
        file_limit=None,
        figsize=None,
        **kwargs,
    ):
        super().__init__(
            data_dir,
            n_components,
            data_spec=data_spec,
            x_lim=x_lim,
            output_dir=output_dir,
            header=header,
            file_ordering=file_ordering,
            file_limit=file_limit,
            figsize=figsize,
            **kwargs,
        )

    def _decomposition(self, *args, **kwargs):
        return iterative_decomposition(*args, **kwargs)


# TODO: Implement RemoteDispatcher and Accumulator agents. See
#  https://github.com/NSLS-II-XPD/ae-gpcam/blob/main/companion/agent.py
